@@ -93,10 +93,10 @@ Some generic base classes and functions can be found in L{zim.utils}
 
 
 # Bunch of meta data, used at least in the about dialog
-__version__ = '0.75.1'
+__version__ = '0.75.2'
 __url__ = 'https://www.zim-wiki.org'
 __author__ = 'Jaap Karssenberg <jaap.karssenberg@gmail.com>'
-__copyright__ = 'Copyright 2008 - 2019 Jaap Karssenberg <jaap.karssenberg@gmail.com>'
+__copyright__ = 'Copyright 2008 - 2023 Jaap Karssenberg <jaap.karssenberg@gmail.com>'
 __license__ = '''\
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -141,7 +141,7 @@ except locale.Error:
 
 _pref_enc = locale.getpreferredencoding()
 if _pref_enc in ('ascii', 'us-ascii', 'ANSI_X3.4-1968'):
-	logger.warn(
+	logger.warning(
 		'Your system encoding is set to %s, if you want support for special characters\n'
 		'or see errors due to encoding, please ensure to configure your system to use "UTF-8"' % _pref_enc
 	)
