@@ -30,16 +30,11 @@ import types
 import ast
 import json
 
-try:
-	import collections.abc as abc
-except ImportError:
-	# python < version 3.3
-	import collections as abc
+import collections.abc as abc
 
-from zim.signals import SignalEmitter, ConnectorMixin, SIGNAL_NORMAL
-from zim.utils import OrderedDict
-from zim.fs import File, FileNotFoundError
-from zim.newfs import FileNotFoundError as NewFileNotFoundError
+from zim.signals import SignalEmitter, ConnectorMixin, SIGNAL_NORMAL, init_signals_for_new_object
+from zim.utils import DefinitionOrderedDict
+from zim.newfs import FileNotFoundError
 from zim.errors import Error
 
 from .basedirs import XDG_CONFIG_HOME
@@ -48,13 +43,8 @@ from .basedirs import XDG_CONFIG_HOME
 logger = logging.getLogger('zim.config')
 
 
-class _MyMeta(type(SignalEmitter), type(OrderedDict)):
-	# Combine meta classes to resolve conflict
-	pass
-
-
-class ControlledDict(OrderedDict, SignalEmitter, ConnectorMixin, metaclass=_MyMeta):
-	'''Sub-class of C{OrderedDict} that tracks modified state.
+class ControlledDict(DefinitionOrderedDict, SignalEmitter, ConnectorMixin):
+	'''Sub-class of C{DefinitionOrderedDict} that tracks modified state.
 	This modified state is recursive for nested C{ControlledDict}s.
 
 	Used as base class for L{SectionedConfigDict} and L{ConfigDict}.
@@ -67,28 +57,34 @@ class ControlledDict(OrderedDict, SignalEmitter, ConnectorMixin, metaclass=_MyMe
 		'changed': (SIGNAL_NORMAL, None, ())
 	}
 
-	def __init__(self, E=None, **F):
-		OrderedDict.__init__(self, E, **F)
+	def __init__(self, E=(), **F):
+		init_signals_for_new_object(self) # Hack, probably needed because we inherit from "dict"
+		DefinitionOrderedDict.__init__(self, E or F)
 		self._modified = False
 
-	# Note that OrderedDict optimizes __getitem__, cannot overload it
-
 	def __setitem__(self, k, v):
-		OrderedDict.__setitem__(self, k, v)
+		DefinitionOrderedDict.__setitem__(self, k, v)
 		if isinstance(v, ControlledDict):
 			self.connectto(v, 'changed', self.on_child_changed)
 		self.emit('changed')
 
 	def __delitem__(self, k):
-		v = OrderedDict.__delitem__(self, k)
-		if isinstance(v, OrderedDict):
+		v = DefinitionOrderedDict.__delitem__(self, k)
+		if isinstance(v, DefinitionOrderedDict):
 			self.disconnect_from(v)
 		self.emit('changed')
 
+	def pop(self, key, default=None):
+		# Only emit changed once here
+		with self.block_signals('changed'):
+			v = DefinitionOrderedDict.pop(self, key, default)
+		self.emit('changed')
+		return v
+
 	def update(self, E=(), **F):
 		# Only emit changed once here
 		with self.block_signals('changed'):
-			OrderedDict.update(self, E, **F)
+			DefinitionOrderedDict.update(self, E, **F)
 		self.emit('changed')
 
 	def changed(self):
@@ -250,7 +246,13 @@ class Boolean(ConfigDefinition):
 class String(ConfigDefinition):
 	'''This class defines a config key that maps to a string'''
 
-	# TODO support esacpe codes \s \t \n \r (see desktop / json spec)
+	# Might expect backslash escaping to be supported here
+	# however we cannot add that backward compatible because this
+	# config type is also used to store paths which may contain
+	# windows paths with unescaped backslashes.
+	# If e.g. a multiline value with escaped "\n" is needed, this
+	# requires a separate "MultiLineString" class or similar.
+	# See also the String definition used in zim.gui.applications
 
 	def __init__(self, default, allow_empty=False):
 		if default == '':
@@ -475,7 +477,7 @@ class ConfigDict(ControlledDict):
 	value does not conform to the definition.
 
 	THis class derives from L{ControlledDict} which in turn derives
-	from L{OrderedDict} so changes to the config can be tracked by the
+	from L{DefinitionOrderedDict} so changes to the config can be tracked by the
 	C{changed} signal, and values are kept in the same order so the order
 	in which items are written to the config file is predictable.
 	'''
@@ -483,7 +485,7 @@ class ConfigDict(ControlledDict):
 	def __init__(self, E=None, **F):
 		assert not (E and F)
 		ControlledDict.__init__(self)
-		self.definitions = OrderedDict()
+		self.definitions = DefinitionOrderedDict()
 		self._input = {}
 		if E or F:
 			self.input(E or F)
@@ -495,10 +497,9 @@ class ConfigDict(ControlledDict):
 		new = self.__class__()
 		new.update(self)
 		new._input.update(self._input)
-		new._keys[:] = list(self._keys)
 		return new
 
-	def update(self, E=None, **F):
+	def update(self, E=(), **F):
 		'''Like C{dict.update()}, copying values from C{E} or C{F}.
 		However if C{E} is also a C{ConfigDict}, also the definitions
 		are copied along.
@@ -517,19 +518,14 @@ class ConfigDict(ControlledDict):
 		keys in the dict, but want to preserve all of them when
 		writing back to a file.
 		'''
-		return dict(self.all_items()) # FIXME should be OrderedDict, but causes test errors
+		return dict(self.all_items()) # FIXME should be DefinitionOrderedDict, but causes test errors
 
 	def all_items(self):
 		# Like items() but returns both defined values and input values
-		for k in self._keys:
-			if k in self._values:
-				yield k, self._values[k]
-			elif k in self._input:
-				yield k, self._input[k]
-			else:
-				pass
-
-	# Note that OrderedDict optimizes __getitem__, cannot overload it
+		for k, v in self.items():
+			yield k, v
+		for k,v in self._input.items():
+			yield k, v
 
 	def __setitem__(self, k, v):
 		if k in self.definitions:
@@ -543,14 +539,10 @@ class ConfigDict(ControlledDict):
 			raise KeyError('Config key "%s" has not been defined' % k)
 
 	def __delitem__(self, k):
-		if k in self._values:
+		if k in self:
 			ControlledDict.__delitem__(self, k)
 		else:
 			del self._input[k]
-			try:
-				self._keys.remove(k)
-			except ValueError:
-				pass
 
 	def input(self, E=None, **F):
 		'''Like C{update()} but won't raise on failures.
@@ -570,8 +562,6 @@ class ConfigDict(ControlledDict):
 				self._set_input(key, value)
 			else:
 				self._input[key] = value # validated later
-				if key not in self._keys:
-					self._keys.append(key)
 
 	def define(self, E=None, **F):
 		'''Set one or more definitions for this config dict
@@ -599,20 +589,26 @@ class ConfigDict(ControlledDict):
 				self._set_input(key, value)
 			else:
 				with self.block_signals('changed'):
-					OrderedDict.__setitem__(self, key, definition.default)
+					DefinitionOrderedDict.__setitem__(self, key, definition.default)
 
 	def _set_input(self, key, value):
 		try:
 			value = self.definitions[key].check(value)
 		except ValueError as error:
-			logger.warn(
-				'Invalid config value for %s: "%s" - %s',
-					key, value, error.args[0]
-			)
+			if error.args:
+				logger.warning(
+					'Invalid config value for %s: "%s" - %s',
+						key, value, error.args[0]
+				)
+			else:
+				logger.warning(
+					'Invalid config value for %s: "%s"',
+						key, value
+				)
 			value = self.definitions[key].default
 
 		with self.block_signals('changed'):
-			OrderedDict.__setitem__(self, key, value)
+			DefinitionOrderedDict.__setitem__(self, key, value)
 
 	def setdefault(self, key, default, check=None, allow_empty=False):
 		'''Set the default value for a configuration item.
@@ -759,7 +755,7 @@ class INIConfigFile(SectionedConfigDict):
 			with self.block_signals('changed'):
 				self.read()
 			self.set_modified(False)
-		except (FileNotFoundError, NewFileNotFoundError):
+		except FileNotFoundError:
 			pass
 
 		if monitor:
@@ -770,7 +766,7 @@ class INIConfigFile(SectionedConfigDict):
 			try:
 				with self.block_signals('changed'):
 					self.read()
-			except (FileNotFoundError, NewFileNotFoundError):
+			except FileNotFoundError:
 				pass
 			else:
 				# First emit top level to allow general changes
@@ -815,12 +811,12 @@ class INIConfigFile(SectionedConfigDict):
 				section = self[name]
 			elif '=' in line:
 				if section is None:
-					logger.warn('Parameter outside section: %s', line)
+					logger.warning('Parameter outside section: %s', line)
 				else:
 					key, string = line.split('=', 1)
 					values.append((str(key.rstrip()), string.lstrip())) # key is not unicode
 			else:
-				logger.warn('Could not parse line: %s', line)
+				logger.warning('Could not parse line: %s', line)
 		else:
 			if values:
 				section.input(values)
