@@ -15,22 +15,37 @@ user can define a new command on the fly.
 '''
 
 import os
+import sys
+import re
 import logging
 from gi.repository import Gtk
 from gi.repository import GObject
 from gi.repository import GLib
 from gi.repository import GdkPixbuf
 
-import zim.fs
-from zim.fs import File, Dir, TmpFile, cleanup_filename
-from zim.config import XDG_DATA_HOME, XDG_DATA_DIRS, data_dirs, SectionedConfigDict, INIConfigFile
-from zim.parsing import split_quoted_strings, uri_scheme
-from zim.applications import Application, WebBrowser, StartFile
-from zim.gui.widgets import Dialog, ErrorDialog, MessageDialog, strip_boolean_result
+from zim.fs import adapt_from_oldfs
+
+from zim.errors import Error
+from zim.newfs import File, LocalFile, TmpFile, FileNotFoundError, \
+	cleanup_filename, get_mimetype_from_path
+from zim.config import XDG_CONFIG_HOME, XDG_CONFIG_DIRS, XDG_DATA_HOME, XDG_DATA_DIRS, \
+	data_dirs, SectionedConfigDict, INIConfigFile, Boolean, ConfigManager
+from zim.parsing import uri_scheme, is_win32_share_re, \
+	normalize_win32_share, is_win32_share_re, is_url_re, is_uri_re, is_www_link_re
+from zim.applications import Application, WebBrowser, StartFile, split_quoted_strings, ApplicationLookUpError
+from zim.gui.widgets import Dialog, ErrorDialog, MessageDialog, QuestionDialog, strip_boolean_result
+
+
 
 
 logger = logging.getLogger('zim.gui.applications')
 
+ui_preferences = (
+	# key, type, category, label, default
+	('always_create_missing_dirs', 'bool', 'Interface',
+		_('Create missing directories automatically\n(If disabled you will be prompted)'), True),
+			# T: option in preferences dialog
+)
 
 def _application_file(path, dirs):
 	# Some logic to check multiple options, e.g. a path of kde-foo.desktop
@@ -55,28 +70,24 @@ def _application_dirs():
 	for dir in data_dirs('applications'):
 		yield dir
 
-	yield XDG_DATA_HOME.subdir('applications')
+	yield XDG_DATA_HOME.folder('applications')
 
 	for dir in XDG_DATA_DIRS:
-		yield dir.subdir('applications')
+		yield dir.folder('applications')
 
 
-def _create_application(dir, Name, Exec, klass=None, NoDisplay=True, **param):
-	n = cleanup_filename(Name.lower()) + '-usercreated'
-	key = n
-	file = dir.file(key + '.desktop')
+def _create_application(dir, basename, Name, Exec, NoDisplay=True, **param):
+	file = dir.file(basename)
 	i = 0
 	while file.exists():
 		assert i < 1000, 'BUG: Infinite loop ?'
 		i += 1
-		key = n + '-' + str(i)
-		file = dir.file(key + '.desktop')
+		basename = basename[:-8] + '-' + str(i) + '.desktop'
+		file = dir.file(basename)
 
-	if klass is None:
-		klass = DesktopEntryFile
-	entry = klass(file)
+	entry = DesktopEntryFile(file)
 	entry.update(
-		Type=param.pop('Type', 'Application'),
+		Type='Application',
 		Version=1.0,
 		NoDisplay=NoDisplay,
 		Name=Name,
@@ -88,22 +99,15 @@ def _create_application(dir, Name, Exec, klass=None, NoDisplay=True, **param):
 	entry.write()
 
 	if param.get('MimeType'):
+		mimetype = param.get('MimeType')
+
+		# Update mimeapps
+		defaults = XDG_CONFIG_HOME.file('mimeapps.list')
+		_update_mimeapps_file(defaults, '[Added Associations]', mimetype, basename, replace=False)
+
 		# Update mimetype cache
 		cache = dir.file('mimeinfo.cache')
-		if not cache.exists():
-			lines = ['[MIME Cache]\n']
-		else:
-			lines = cache.readlines()
-
-		mimetype = param.get('MimeType')
-		for i, line in enumerate(lines):
-			if line.startswith(mimetype + '='):
-				lines[i] = line.strip() + ';' + key + '.desktop\n'
-				break
-		else:
-			lines.append(mimetype + '=' + key + '.desktop\n')
-
-		cache.writelines(lines)
+		_update_mimeapps_file(cache, '[MIME Cache]', mimetype, basename, replace=False)
 
 	return entry
 
@@ -114,17 +118,20 @@ def get_mimetype(obj):
 	@param obj: a L{File} object, or an URL
 	@returns: mimetype or C{None}
 	'''
-
-	for method in ('get_mimetype', 'mimetype'): # zim.fs.File, newfs
-		if hasattr(obj, method):
-			return getattr(obj, method)()
+	if hasattr(obj, 'mimetype'):
+		return obj.mimetype()
 	else:
 		scheme = uri_scheme(obj)
 		if scheme in (None, 'file'):
 			try:
-				return File(obj).get_mimetype()
+				return get_mimetype_from_path(FilePath(obj).uri)
 			except:
-				return None
+				try:
+					# Mal-formed or relative path, still valid mimetype from extension
+					return get_mimetype_from_path(obj)
+				except:
+					logger.exception('Failed mimetype for %s')
+					return None
 		else:
 			return "x-scheme-handler/%s" % scheme
 
@@ -211,6 +218,49 @@ def _read_comment_from(file):
 		else:
 			return None
 
+def _update_mimeapps_file(file, section, key, value, replace):
+	assert section.startswith('[')
+	assert not (not value and not replace)
+	assert value is None or value.endswith('.desktop')
+	key = key + '='
+
+	if file.exists():
+		lines = file.readlines()
+	else:
+		lines = [section + '\n']
+
+	insert_location = None
+	in_section = False
+	for i, line in enumerate(lines):
+		if line.startswith(section):
+			in_section = True
+			insert_location = i + 1
+		elif in_section:
+			if line.startswith(key):
+				if replace:
+					if value:
+						lines[i] = key + value + '\n'
+					else:
+						lines[i] = ''
+				else:
+					lines[i] = key + value + ';' + lines[i][len(key):]
+				break
+			elif line.startswith('['):
+				in_section = False
+			elif not line.isspace():
+				insert_location = i + 1
+			else:
+				pass # empty lines
+	else:
+		if value:
+			if insert_location is None: # section not found - add it
+				lines.extend(['\n', section + '\n', key + value + '\n'])
+			else:
+				lines.insert(insert_location, key + value + '\n')
+
+	file.writelines(lines)
+
+
 
 class ApplicationManager(object):
 	'''Manager object for dealing with desktop applications. Uses the
@@ -218,6 +268,8 @@ class ApplicationManager(object):
 	installed applications.
 	'''
 
+	_defaults_app_cache = {}
+
 	@staticmethod
 	def get_application(name):
 		'''Get an application by name. Will search installed ".desktop"
@@ -234,7 +286,7 @@ class ApplicationManager(object):
 
 		file = _application_file(key, _application_dirs())
 		if file:
-			return DesktopEntryFile(File(file))
+			return DesktopEntryFile(LocalFile(file))
 		elif name == 'webbrowser':
 			return WebBrowser()
 		elif name == 'startfile':
@@ -249,30 +301,61 @@ class ApplicationManager(object):
 		@param mimetype: the mime-type of the file (e.g. "text/html")
 		@returns: an L{Application} object or C{None}
 		'''
-		## Based on logic from xdg-mime defapp_generic()
-		## Obtained from http://portland.freedesktop.org/wiki/ (2012-05-31)
+		# TODO: also check timestamp of mimeapps for validity of the cache ?
+		if mimetype in klass._defaults_app_cache \
+			and klass._defaults_app_cache[mimetype] is not None \
+				and not klass._defaults_app_cache[mimetype].file.exists():
+					del klass._defaults_app_cache[mimetype]
+
+		if mimetype not in klass._defaults_app_cache:
+			klass._defaults_app_cache[mimetype] = klass._get_default_application(mimetype)
+
+		return klass._defaults_app_cache[mimetype]
+
+	@staticmethod
+	def _mimeapps_files():
+		desktops = []
+		if os.environ.get('XDG_CURRENT_DESKTOP'):
+			desktops = os.environ['XDG_CURRENT_DESKTOP'].split(';')
+		folders = [XDG_CONFIG_HOME] + XDG_CONFIG_DIRS
+		folders += [f.folder('applications') for f in [XDG_DATA_HOME] + XDG_DATA_DIRS]
+
+		for folder in folders:
+			for desktop in desktops:
+				file = folder.file('%s-mimeapps.list' % desktop.lower())
+				if file.exists():
+					yield file
+
+			file = folder.file('mimeapps.list')
+			if file.exists():
+				yield file
+
+		for folder in _application_dirs():
+			file = folder.file('defaults.list')
+			if file.exists():
+				yield file
+
+	@classmethod
+	def _get_default_application(klass, mimetype):
+		## Based on https://specifications.freedesktop.org/mime-apps-spec/latest/
+		## Version 1.0 dated 2 April 2014
 		##
-		## Considered calling xdg-mime directly with code below as fallback.
-		## But xdg-mime has only a special case for KDE, all others are generic.
-		## Our purpose is to be able to set defaults ourselves and read them back.
-		## If we fail we fallback to opening files with the file browser which
-		## defaults to xdg-open. So even if the system does not support the
-		## generic implementation, it will behave sanely and fall back to system
-		## defaults.
-
-		## TODO: optimize for being called very often ?
-
-		for dir in _application_dirs():
-			default_file = dir.file('defaults.list')
-			if not default_file.exists():
-				continue
+		## Implemented from scratch to ensure platform independent support
+		##
+		## Kept "defaults.list" file support for backward compatibility
+		## when changing zim versions.
 
+		for default_file in klass._mimeapps_files():
+			in_default_section = False
 			for line in default_file.readlines():
-				if line.startswith(mimetype + '='):
+				if line.startswith('[Default Applications]'):
+					in_default_section = True
+				elif in_default_section and line.startswith('['):
+					break # new group, look no further
+
+				if in_default_section and line.startswith(mimetype + '='):
 					_, key = line.strip().split('=', 1)
 					for k in key.split(';'):
-						# Copied logic from xdg-mime, apparently entries
-						# can be ";" seperated lists
 						k = k.strip()
 						application = klass.get_application(k)
 						if application is not None:
@@ -281,8 +364,8 @@ class ApplicationManager(object):
 		else:
 			return None
 
-	@staticmethod
-	def set_default_application(mimetype, application):
+	@classmethod
+	def set_default_application(klass, mimetype, application):
 		'''Set the default application to open a file with a specific
 		mimetype. Updates the C{applications/defaults.list} file.
 		As a special case when you set the default to C{None} it will
@@ -291,10 +374,10 @@ class ApplicationManager(object):
 		@param mimetype: the mime-type of the file (e.g. "text/html")
 		@param application: an L{Application} object or C{None}
 		'''
-		## Based on logic from xdg-mime make_default_generic()
-		## Obtained from http://portland.freedesktop.org/wiki/ (2012-05-31)
-		##
-		## See comment in get_default_application()
+		## Based on https://specifications.freedesktop.org/mime-apps-spec/latest/
+		## Version 1.0 dated 2 April 2014
+
+		klass._defaults_app_cache[mimetype] = application
 
 		if application is not None:
 			if not isinstance(application, str):
@@ -303,16 +386,8 @@ class ApplicationManager(object):
 			if not application.endswith('.desktop'):
 				application += '.desktop'
 
-		default_file = XDG_DATA_HOME.file('applications/defaults.list')
-		if default_file.exists():
-			lines = default_file.readlines()
-			lines = [l for l in lines if not l.startswith(mimetype + '=')]
-		else:
-			lines = ['[Default Applications]\n']
-
-		if application:
-			lines.append('%s=%s\n' % (mimetype, application))
-		default_file.writelines(lines)
+		file = XDG_CONFIG_HOME.file('mimeapps.list')
+		_update_mimeapps_file(file, '[Default Applications]', mimetype, application, replace=True)
 
 	@staticmethod
 	def create(mimetype, Name, Exec, **param):
@@ -333,9 +408,10 @@ class ApplicationManager(object):
 		@returns: the L{DesktopEntryFile} object with some
 		sensible defaults for a user created application entry.
 		'''
-		dir = XDG_DATA_HOME.subdir('applications')
+		dir = XDG_DATA_HOME.folder('applications')
 		param['MimeType'] = mimetype
-		file = _create_application(dir, Name, Exec, **param)
+		basename = cleanup_filename(Name.lower()) + '-usercreated.desktop'
+		file = _create_application(dir, basename, Name, Exec, **param)
 		return file
 
 	@classmethod
@@ -344,7 +420,7 @@ class ApplicationManager(object):
 		# should handle all file types
 		if os.name == 'nt':
 			return StartFile()
-		elif os.name == 'darwin':
+		elif sys.platform == 'darwin':
 			app = Application('open')
 		else: # linux and friends
 			app = Application('xdg-open')
@@ -363,7 +439,7 @@ class ApplicationManager(object):
 		# Don't use mimetype lookup here, this is a fallback
 		if os.name == 'nt':
 			return StartFile()
-		elif os.name == 'darwin':
+		elif sys.platform == 'darwin':
 			app = Application('open')
 		else: # linux and friends
 			app = Application('xdg-email')
@@ -395,30 +471,55 @@ class ApplicationManager(object):
 		@returns: a list of L{Application} objects that are known to
 		be able to handle this file type
 		'''
+		## Supporting mimapps.list
+		## Based on https://specifications.freedesktop.org/mime-apps-spec/latest/
+		## Version 1.0 dated 2 April 2014
+		##
+		## Also using older "mimeinfo.cache" which is deprecated in the spec
+		## but still in use at least on Ubuntu as of Jan 2021
+		##
+		## TODO: cache these as well ?
+
 		seen = set()
 		entries = []
+		blacklist = set()
 		key = '%s=' % mimetype
+
+		def add_entry(basename, dirs):
+			if basename not in seen and basename not in blacklist:
+				file = _application_file(basename, dirs)
+				if file:
+					entries.append(DesktopEntryFile(LocalFile(file)))
+					seen.add(basename)
+
+		for file in klass._mimeapps_files():
+			section=None
+			for line in file.readlines():
+				if line.startswith('['):
+					section=line.strip()
+				elif line.startswith(key):
+					if section in ('[Default Applications]', '[Added Associations]'):
+						for basename in [b for b in line[len(key):].strip().split(';') if b]:
+							add_entry(basename, _application_dirs())
+					elif section == '[Removed Associations]':
+						for basename in [b for b in line[len(key):].strip().split(';') if b]:
+							blacklist.add(basename)
+
 		for dir in _application_dirs():
 			cache = dir.file('mimeinfo.cache')
 			if not cache.exists():
 				continue
 			for line in cache.readlines():
 				if line.startswith(key):
-					for basename in line[len(key):].strip().split(';'):
-						if basename in seen:
-							continue
-						else:
-							file = _application_file(basename, (dir,))
-							if file:
-								entries.append(DesktopEntryFile(File(file)))
-								seen.add(basename)
+					for basename in [b for b in line[len(key):].strip().split(';') if b]:
+						add_entry(basename, (dir,))
 
 		if mimetype in ('x-scheme-handler/http', 'x-scheme-handler/https'):
 			# Since "x-scheme-handler" is not in the standard, some browsers
 			# only identify themselves with "text/html".
 			for entry in klass.list_applications('text/html', nodisplay): # recurs
 				basename = entry.key + '.desktop'
-				if not basename in seen:
+				if basename not in seen and basename not in blacklist:
 					entries.append(entry)
 					seen.add(basename)
 
@@ -428,15 +529,6 @@ class ApplicationManager(object):
 		return entries
 
 
-from zim.errors import Error
-from zim.parsing import is_win32_share_re, is_url_re, is_uri_re
-
-from zim.fs import adapt_from_newfs, normalize_win32_share
-from zim.newfs import FileNotFoundError
-
-from zim.gui.widgets import QuestionDialog
-
-
 class NoApplicationFoundError(Error):
 	'''Exception raised when an application was not found'''
 	pass
@@ -460,22 +552,18 @@ def open_file(widget, file, mimetype=None, callback=None):
 	ignore the specified mimetype)
 	'''
 	logger.debug('open_file(%s, %s)', file, mimetype)
-	file = adapt_from_newfs(file)
-	assert isinstance(file, (File, Dir))
-	if isinstance(file, (File)) and file.isdir():
-		file = Dir(file.path)
-
+	file = adapt_from_oldfs(file)
 	if not file.exists():
 		raise FileNotFoundError(file)
 
-	if isinstance(file, File): # File
+	if isinstance(file, File):
 		manager = ApplicationManager()
 		if mimetype is None:
-			entry = manager.get_default_application(file.get_mimetype())
+			entry = manager.get_default_application(file.mimetype())
 			if entry is not None:
 				_open_with(widget, entry, file, callback)
 			else:
-				_open_with_filebrowser(widget, file, callback)
+				_open_with_filebrowser(widget, file.uri, callback)
 
 		else:
 			entry = manager.get_default_application(mimetype)
@@ -485,7 +573,7 @@ def open_file(widget, file, mimetype=None, callback=None):
 				raise NoApplicationFoundError('No Application found for: %s' % mimetype)
 				# Do not go to fallback, we can not force
 				# mimetype for fallback
-	else: # Dir
+	else: # Folder
 		_open_with_filebrowser(widget, file, callback)
 
 
@@ -497,12 +585,20 @@ def open_folder_prompt_create(widget, folder):
 	try:
 		open_folder(widget, folder)
 	except FileNotFoundError:
-		if QuestionDialog(widget, (
-			_('Create folder?'),
-				# T: Heading in a question dialog for creating a folder
-			_('The folder "%s" does not yet exist.\nDo you want to create it now?') % folder.basename
-				# T: Text in a question dialog for creating a folder, %s will be the folder base name
-		)).run():
+		config = ConfigManager.preferences['Application']
+		config.setdefault('always_create_missing_dirs', False)
+		create_dir = config['always_create_missing_dirs']
+
+		# prompt user if we're not asked to always create directories
+		if not create_dir:
+			create_dir = QuestionDialog(widget, (
+				_('Create folder?'),
+					# T: Heading in a question dialog for creating a folder
+				_('The folder "%s" does not yet exist.\nDo you want to create it now?') % folder.basename
+					# T: Text in a question dialog for creating a folder, %s will be the folder base name
+				)).run()
+
+		if create_dir:
 			folder.touch()
 			open_folder(widget, folder)
 
@@ -515,8 +611,7 @@ def open_folder(widget, folder):
 	see L{open_folder_prompt_create} for alternative behavior when folder
 	does not exist.
 	'''
-	dir = adapt_from_newfs(folder)
-	open_file(widget, dir)
+	open_file(widget, folder)
 
 
 def open_url(widget, url):
@@ -535,6 +630,8 @@ def open_url(widget, url):
 		if os.name == 'nt':
 			return _open_with_filebrowser(widget, url)
 		# else consider as a x-scheme-handler/smb type URI below
+	elif is_www_link_re.match(url):
+		url = 'http://' + url
 	elif not is_uri_re.match(url):
 		raise ValueError('Not an URL: %s' % url)
 	else:
@@ -549,7 +646,6 @@ def open_url(widget, url):
 		# Special case for outlook folder paths on windows
 		os.startfile(url)
 	else:
-		from zim.gui.applications import get_mimetype
 		manager = ApplicationManager()
 		type = get_mimetype(url) # Supports "x-scheme-handler/... for URL schemes"
 		logger.debug('Got type "%s" for "%s"', type, url)
@@ -577,9 +673,9 @@ def _open_with(widget, entry, uri, callback=None):
 		entry.spawn((uri,)) # E.g. webbrowser module does not support callback
 
 
-def _open_with_filebrowser(widget, file, callback=None):
+def _open_with_filebrowser(widget, uri, callback=None):
 	entry = ApplicationManager.get_fallback_filebrowser()
-	_open_with(widget, entry, file, callback)
+	_open_with(widget, entry, uri, callback)
 
 
 def _open_with_emailclient(widget, uri):
@@ -666,11 +762,35 @@ from zim.config import String as BaseString
 from zim.config import Boolean as BaseBoolean
 from zim.config import Float as Numeric
 
-class String(BaseString):
+
+class XDGString(BaseString):
+
+	def check(self, value):
+		if isinstance(value, str):
+			value = BaseString.check(self, value)
+			if isinstance(value, str):
+				# According to the XDG Desktop entry spec we should support
+				# the follwoing codes: \s, \n, \t, \r, and \\
+				# However, only do \\ to avoid conflicts invalid strings
+				# also no use case for other codes in keys we use (?)
+				return re.sub(r'\\\\', '\\\\', value)
+			else:
+				return value
+		else:
+			return BaseString.check(self, value)
+
+	def tostring(self, value):
+		if value is None:
+			return ''
+		else:
+			return value.replace('\\', '\\\\')
+
+
+class String(XDGString):
 
 	def check(self, value):
 		# Only ascii chars allowed in these keys
-		value = BaseString.check(self, value)
+		value = XDGString.check(self, value)
 		if isinstance(value, str):
 			try:
 				x = value.encode('ascii')
@@ -681,7 +801,7 @@ class String(BaseString):
 		return value
 
 
-class LocaleString(BaseString):
+class LocaleString(XDGString):
 	pass # utf8 already supported by default
 
 
@@ -689,7 +809,7 @@ class IconString(LocaleString):
 
 	def check(self, value):
 		if hasattr(value, 'path'):
-			return value.path  # prevent fallback via serialize_zim_config to user_path
+			return value.path  # prevent fallback via serialize_zim_config to userpath
 		else:
 			return LocaleString.check(self, value)
 
@@ -701,6 +821,31 @@ class Boolean(BaseBoolean):
 		return str(value).lower()
 
 
+class TerminalLookUpError(ApplicationLookUpError):
+
+	def __init__(self, cmd):
+		self.msg = _('Cound not find terminal emulator to run application: %s') % cmd
+			# T: Error message when external application could not be run in terminal, %s is the command
+
+
+_terminal_commands = (
+	('xdg-terminal'),
+	#('x-terminal-emulator', '-x'), # Testing shows "-x" does not work as advertised :(
+	('gnome-terminal', '-x'),
+	('xterm', '-e'),
+)
+
+def _get_terminal_command(appcmd):
+	# Intended as *minimal* implementation for getting a terminal emulator.
+	# Since there is no spec how to do this lookup, don't add any complexity
+	# here other than maybe a switch per OS. Instead pull-in xdg-terminal as
+	# dependency if more sophisticated logic is needed.
+	for cmd in _terminal_commands:
+		if Application(cmd).tryexec():
+			return cmd + tuple(appcmd)
+	else:
+		raise TerminalLookUpError(appcmd[0])
+
 
 class DesktopEntryDict(SectionedConfigDict, Application):
 	'''Base class for L{DesktopEntryFile}, defines most of the logic.
@@ -787,6 +932,10 @@ class DesktopEntryDict(SectionedConfigDict, Application):
 	def nodisplay(self):
 		return self['Desktop Entry'].get('NoDisplay', False)
 
+	@property
+	def needsterminal(self):
+		return self['Desktop Entry'].get('Terminal', False)
+
 	@property
 	def tryexeccmd(self):
 		return self['Desktop Entry'].get('TryExec')
@@ -804,8 +953,7 @@ class DesktopEntryDict(SectionedConfigDict, Application):
 		if not icon:
 			return None
 
-		if isinstance(icon, File):
-			icon = icon.path
+		icon = icon.path if hasattr(icon, 'path') else icon
 
 		w, h = strip_boolean_result(Gtk.icon_size_lookup(size))
 
@@ -829,71 +977,80 @@ class DesktopEntryDict(SectionedConfigDict, Application):
 		@param args: list of either URLs or L{File} objects
 		@returns: the full command to execute as a tuple
 		'''
+		# The XDG Desktop Entry spec specifies replacement of field codes when
+		# they appear as a single argument. Handling of field codes inside
+		# arguments is explicitly left undefined. Users seem to expect these
+		# to work, so at least interpolate the values that expand to a single
+		# argument.
+
 		assert args is None or isinstance(args, (list, tuple))
 
 		def uris(args):
 			uris = []
 			for arg in args:
-				if isinstance(arg, (File, Dir)):
+				if hasattr(arg, 'uri'):
 					uris.append(arg.uri)
 				else:
 					uris.append(str(arg))
 			return uris
 
-		cmd = split_quoted_strings(self['Desktop Entry']['Exec'])
-		if args is None or len(args) == 0:
-			if '%f' in cmd:
-				cmd.remove('%f')
-			elif '%F' in cmd:
-				cmd.remove('%F')
-			elif '%u' in cmd:
-				cmd.remove('%u')
-			elif '%U' in cmd:
-				cmd.remove('%U')
-		elif '%f' in cmd:
-			assert len(args) == 1, 'application takes one file name'
-			i = cmd.index('%f')
-			cmd[i] = str(args[0])
-		elif '%F' in cmd:
-			i = cmd.index('%F')
-			for arg in reversed(list(map(str, args))):
-				cmd.insert(i, str(arg))
-			cmd.remove('%F')
-		elif '%u' in cmd:
-			assert len(args) == 1, 'application takes one url'
-			i = cmd.index('%u')
-			cmd[i] = uris(args)[0]
-		elif '%U' in cmd:
-			i = cmd.index('%U')
-			for arg in reversed(uris(args)):
-				cmd.insert(i, str(arg))
-			cmd.remove('%U')
-		else:
-			cmd.extend(list(map(str, args)))
+		cmd = []
+		seen_arg_code = False
 
-		if '%i' in cmd:
-			if 'Icon' in self['Desktop Entry'] \
-			and self['Desktop Entry']['Icon']:
-				i = cmd.index('%i')
-				cmd[i] = self['Desktop Entry']['Icon']
-				cmd.insert(i, '--icon')
+		def sub_field_code(m):
+			nonlocal seen_arg_code
+			m = m.group()
+			if m == '%f':
+				seen_arg_code = True
+				return str(args[0]) if args else ''
+			elif m == '%u':
+				seen_arg_code = True
+				return uris(args)[0] if args else ''
+			elif m == '%k':
+				return self.file.path if hasattr(self, 'file') else ''
+			elif m == '%c':
+				return self.name
 			else:
-				cmd.remove('%i')
+				return '%'
 
-		if '%c' in cmd:
-			i = cmd.index('%c')
-			cmd[i] = self.name
-
-		if '%k' in cmd:
-			i = cmd.index('%k')
-			if hasattr(self, 'file'):
-				cmd[i] = self.file.path
+		for word in split_quoted_strings(self['Desktop Entry']['Exec']):
+			# These expnd to multiple arguments and cannot be interpolated
+			if word in ('%d', '%D', '%n', '%N', '%v', '%m'):
+				continue # deprecated codes
+			elif word == '%F':
+				if args:
+					cmd.extend([str(a) for a in args])
+				seen_arg_code = True
+				continue
+			elif word == '%U':
+				if args:
+					cmd.extend([str(a) for a in uris(args)])
+				seen_arg_code = True
+				continue
+			elif word == '%i':
+				if 'Icon' in self['Desktop Entry'] \
+				and self['Desktop Entry']['Icon']:
+					cmd.extend(['--icon', self['Desktop Entry']['Icon']])
+				continue
+			elif word in ('%f', '%u') and not args:
+				seen_arg_code = True
+				continue
 			else:
-				cmd[i] = ''
+				# These can be interpolated inside arguments
+				word = re.sub('%%|%[fukc]', sub_field_code, word)
+				cmd.append(word)
+
+		if not seen_arg_code and args:
+			cmd.extend([str(a) for a in args])
 
 		return tuple(cmd)
 
-	_cmd = parse_exec # To hook into Application.spawn and Application.run
+	def _cmd(self, args):
+		# Hook into Application.spawn and Application.run
+		cmd = self.parse_exec(args)
+		if self.needsterminal:
+			cmd = _get_terminal_command(cmd)
+		return cmd
 
 	def update(self, E=(), **F):
 		'''Same as C{dict.update()}'''
